'''Raw socket layer.

Mostly inspired from standard lib `asyncio`.
'''

import os
import socket

from errno import ENOTCONN, ECONNABORTED
from typing import List, Tuple, Union

from pymaid.core import get_running_loop, run_in_threadpool

HAS_IPv6_FAMILY = hasattr(socket, 'AF_INET6')
HAS_IPv6_PROTOCOL = hasattr(socket, 'IPPROTO_IPV6')


async def sock_connect(address: Tuple[str, int]) -> socket.socket:
    if isinstance(address, str):
        infos = [(socket.AF_UNIX, socket.SOCK_STREAM, 0, '', address)]
    else:
        host, port = address
        infos = await run_in_threadpool(
            socket.getaddrinfo, args=(host, port, 0, socket.SOCK_STREAM),
        )
    loop = get_running_loop()
    err = None
    for res in infos:
        af, socktype, proto, canonname, sa = res
        sock = None
        retried = False
        while 1:
            try:
                sock = socket.socket(af, socktype, proto)
                sock.setblocking(False)
                await loop.sock_connect(sock, sa)
                if socktype == socket.SOCK_STREAM and af != socket.AF_UNIX:
                    sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                # NOTE:
                # when doing a lots connect to remote side under heavy pressure
                # it would sometimes getting ENOTCONN when call getpeername
                # check it here, if occured, raise it to retry
                # *why* sock_connect above does not handle this case?
                sock.getpeername()
                # Break explicitly a reference cycle
                err = None
                return sock
            except socket.error as _:
                if _.errno in {ECONNABORTED, ENOTCONN} and not retried:
                    retried = True
                    continue
                err = _
                if sock is not None:
                    sock.close()
                break

    if err is not None:
        try:
            raise err
        finally:
            # Break explicitly a reference cycle
            err = None
    else:
        raise socket.error('getaddrinfo returns an empty list')


async def sock_listen(
    address: Union[Tuple[str, int], str],
    family: socket.AddressFamily = socket.AF_UNSPEC,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    backlog: int = 128,
    reuse_address: bool = True,
    reuse_port: bool = False,
) -> List[socket.socket]:
    '''Create sockets listening on `address`.

    The address parameter can be a string, in that case the TCP server is
    bound to unix domain sock.

    The address parameter can also be a tuple of string and int, in that case
    the TCP server is bound to host and port.
    If a host appears multiple times (possibly indirectly e.g. when hostnames
    resolve to the same IP address), the server is only bound once to that
    host.

    Return a Server object which can be used to stop the service.

    This is a coroutine.
    '''
    sockets = []

    if isinstance(address, str):
        if os.path.exists(address):
            os.unlink(address)
        infos = [(socket.AF_UNIX, socket.SOCK_STREAM, 0, '', address)]
    else:
        host, port = address
        infos = await run_in_threadpool(
            socket.getaddrinfo,
            args=(host, port, family, socket.SOCK_STREAM),
            kwargs={'flags': flags},
        )
        infos = set(infos)
    try:
        for res in infos:
            af, socktype, proto, canonname, sa = res
            try:
                sock = socket.socket(af, socktype, proto)
            except socket.error:
                # Assume it's a bad family/type/protocol combination.
                continue
            sock.setblocking(False)
            if socktype == socket.SOCK_STREAM and af != socket.AF_UNIX:
                sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            if reuse_address:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
            if reuse_port:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            # Disable IPv4/IPv6 dual stack support (enabled by
            # default on Linux) which makes a single socket
            # listen on both address families.
            if HAS_IPv6_FAMILY and HAS_IPv6_PROTOCOL and af == socket.AF_INET6:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, True)
            try:
                sock.bind(sa)
            except OSError as err:
                raise OSError(
                    err.errno,
                    f'error while binding on address {infos}: {err.strerror}'
                ) from None
            sock.listen(backlog)
            sockets.append(sock)
    except Exception:
        for sock in sockets:
            sock.close()
        raise

    if not sockets:
        raise socket.error('getaddrinfo returns an empty list')
    return sockets



from argparse import ArgumentParser
import re


def get_base_parser():
    parser = ArgumentParser()
    parser.add_argument(
        '--address', type=str, default='/tmp/pymaid.sock',
        help='transport address',
    )
    parser.add_argument(
        '--uvloop', action='store_true', default=False, help='use uvloop'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', default=False,
        help='print debug info'
    )
    return parser


def get_server_parser():
    parser = get_base_parser()
    parser.add_argument(
        '--backlog', type=int, default=1024, help='listen backlog',
    )
    return parser


def get_client_parser():
    parser = get_base_parser()
    parser.add_argument(
        '-c', dest='concurrency', type=int, default=100, help='concurrency',
    )
    parser.add_argument(
        '-r', dest='request', type=int, default=100, help='request per client',
    )
    parser.add_argument('--msize', type=int, default=1024, help='message size')
    parser.add_argument(
        '--timeout', type=int, default=60, help='timeout, in seconds',
    )
    return parser


def parse_args(parser):
    args = parser.parse_args()
    if re.search(r':\d+$', args.address):
        address, port = args.address.split(':')
        args.address = (address, int(port))
    print(args)
    if args.uvloop:
        import uvloop
        uvloop.install()
    if args.verbose:
        args.debug = print
    else:
        def debug(*args, **kwargs):
            pass
        args.debug = debug
    return args



##############################################################################
#
#  Copyright 2011-2013 Tavendo GmbH
#
#  Note:
#
#  This code is a Python implementation of the algorithm
#
#            'Flexible and Economical UTF-8 Decoder'
#
#  by Bjoern Hoehrmann
#
#       bjoern@hoehrmann.de
#       http://bjoern.hoehrmann.de/utf-8/decoder/dfa/
#
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
##############################################################################


# use Cython implementation of UTF8 validator if available
#
try:
    from wsaccel.utf8validator import Utf8Validator
except ImportError:

    # fallback to pure Python implementation
    class Utf8Validator:
        '''Incremental UTF-8 validator with constant memory consumption
        (minimal state).

        Implements the algorithm 'Flexible and Economical UTF-8 Decoder' by
        Bjoern Hoehrmann (http://bjoern.hoehrmann.de/utf-8/decoder/dfa/).
        '''

        # DFA transitions
        UTF8VALIDATOR_DFA = [
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 00..1f
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 20..3f
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 40..5f
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  # 60..7f
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,  # 80..9f
            7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
            7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,  # a0..bf
            8, 8, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
            2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,  # c0..df
            0xa, 0x3, 0x3, 0x3, 0x3, 0x3, 0x3, 0x3,
            0x3, 0x3, 0x3, 0x3, 0x3, 0x4, 0x3, 0x3,  # e0..ef
            0xb, 0x6, 0x6, 0x6, 0x5, 0x8, 0x8, 0x8,
            0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8,  # f0..ff
            0x0, 0x1, 0x2, 0x3, 0x5, 0x8, 0x7, 0x1,
            0x1, 0x1, 0x4, 0x6, 0x1, 0x1, 0x1, 0x1,  # s0..s0
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1,  # s1..s2
            1, 2, 1, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1,  # s3..s4
            1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1, 1, 3, 1, 3, 1, 1, 1, 1, 1, 1,  # s5..s6
            1, 3, 1, 1, 1, 1, 1, 3, 1, 3, 1, 1, 1, 1, 1, 1,
            1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  # s7..s8
        ]

        UTF8_ACCEPT = 0
        UTF8_REJECT = 1

        def __init__(self):
            self.reset()

        def decode(self, b):
            '''Eat one UTF-8 octet, and validate on the fly.

            Returns UTF8_ACCEPT when enough octets have been consumed,
            in which case self.codepoint contains the decoded Unicode code
            point.

            Returns UTF8_REJECT when invalid UTF-8 was encountered.

            Returns some other positive integer when more octets need to be
            eaten.
            '''
            dfa = Utf8Validator.UTF8VALIDATOR_DFA[b]

            state = self.state
            if state != Utf8Validator.UTF8_ACCEPT:
                self.codepoint = (b & 0x3f) | (self.codepoint << 6)
            else:
                self.codepoint = (0xff >> dfa) & b

            state = Utf8Validator.UTF8VALIDATOR_DFA[256 + state * 16 + dfa]
            self.state = state

            return state

        def reset(self):
            '''Reset validator to start new incremental UTF-8 decode/validation.'''
            self.state = Utf8Validator.UTF8_ACCEPT
            self.codepoint = 0
            self.i = 0

        def validate(self, ba):
            '''Incrementally validate a chunk of bytes provided as string.

            Will return a quad
            (valid, endsOnCodePoint, currentIndex, totalIndex)

            As soon as an octet is encountered which renders the octet sequence
            invalid, a quad with valid == False is returned.
            currentIndex returns the index within the currently consumed chunk,
            and totalIndex the index within the total consumed sequence that
            was the point of bail out.
            When valid == True, currentIndex will be len(ba) and totalIndex the
            total amount of consumed bytes.
            '''

            size = len(ba)
            DFA = Utf8Validator.UTF8VALIDATOR_DFA

            for i in range(size):
                # optimized version of decode()
                # since we are not interested in actual code points

                self.state = DFA[256 + (self.state << 4) + DFA[ord(ba[i])]]

                if self.state == Utf8Validator.UTF8_REJECT:
                    self.i += i
                    return False, False, i, self.i

            self.i += size

            return True, self.state == Utf8Validator.UTF8_ACCEPT, size, self.i



'''pymaid separate network into two layer

- transport layer
    transport layer concern about how data transmit
    for example: stream for tcp; packet for udp
    and security of transmission: with or without ssl/tls

- protocol layer
    protocol layer concern about how app data format
    it does not care where the data comes from
    stream, packet, file, or pipeline are not different

    you feed it data, and it gives you protocol messages, that's all

    for example: http, protocol buffer, websocket
    and also has the ability to wrap builtin Protocols
    like AppProtocol(PBProtocol) or even AppProtocol(PBProtocol(WSProtocol))
'''

__all__ = ['Protocol', 'Transport', 'Stream', 'Datagram']

import socket
import ssl as _ssl

from typing import Optional, Tuple, Union


from .channel import ChannelType, StreamChannel
from .stream import Stream


async def dial_stream(
    address: Tuple[str, int],
    *,
    channel_class: ChannelType = StreamChannel,
    transport_class: Stream = Stream,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
):
    '''Create channel can connect to `address`.

    The host parameter can be a string, in that case the TCP Channel is
    connect to unix domain sock.

    The host parameter can also be a tuple of string and int, in that case
    the TCP Channel is connect to the host and port. If a host
    appears multiple times (possibly indirectly e.g. when hostnames
    resolve to the same IP address), the Channel is only connect once to that
    host.

    Return a Channel object which can be used to manage the streams.

    This method is a coroutine.
    '''
    return channel_class(
        address=address,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
    )


async def serve_stream(
    address: Union[Tuple[str, int], str],
    *,
    family: socket.AddressFamily = socket.AF_UNSPEC,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    backlog: int = 128,
    reuse_address: bool = True,
    reuse_port: bool = False,
    ssl_context: Union[None, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    channel_class: ChannelType = StreamChannel,
    transport_class: Stream = Stream,
    start_serving: bool = True,
):
    '''Create channel listening on `address`.

    The host parameter can be a string, in that case the TCP Channel is
    bound to unix domain sock.

    The host parameter can also be a tuple of string and int, in that case
    the TCP Channel is bound to the host and port. If a host
    appears multiple times (possibly indirectly e.g. when hostnames
    resolve to the same IP address), the Channel is only bound once to that
    host.

    Return a Channel object which can be used to manage the streams.

    This method is a coroutine.
    '''
    channel = channel_class(
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
    )
    await channel.listen(
        address,
        family=family,
        flags=flags,
        backlog=backlog,
        reuse_address=reuse_address,
        reuse_port=reuse_port,
    )
    if start_serving:
        channel.start()
    return channel



from unittest import TestCase

from pymaid import error


class ErrorTest(TestCase):

    def setUp(self):
        error.ErrorManager.managers.clear()
        error.ErrorManager.codes.clear()

    def test_base(self):
        self.assertIsNone(error.BaseEx.code)
        self.assertEqual(error.BaseEx.message, 'BaseEx')

        class Error1(error.Error):
            code = 'Error1'
            message = 'Error1: {}'
        self.assertEqual(Error1.__module__, __name__)

    def test_error(self):
        class Error1(error.Error):
            code = 'Error1'
            message = 'Error1: {}'

        self.assertEqual(Error1.code, 'Error1')
        self.assertEqual(Error1.message, 'Error1: {}')

        ex = Error1('Oops', number=1, data={'name': 'Whatever'})
        self.assertEqual(ex.code, 'Error1')
        self.assertEqual(ex.message, 'Error1: Oops')
        self.assertDictEqual(ex.data, {'name': 'Whatever'})
        self.assertEqual(
            str(ex),
            "[ERROR][code|Error1][message|Error1: Oops]"
            "[data|{'name': 'Whatever'}]"
        )

        self.assertIsInstance(ex, error.Error)
        self.assertIsInstance(ex, error.BaseEx)
        self.assertNotIsInstance(ex, error.Warning)

    def test_warning(self):
        class Warning1(error.Warning):
            code = 'Warning1'
            message = 'Warning1: {}'

        self.assertEqual(Warning1.code, 'Warning1')
        self.assertEqual(Warning1.message, 'Warning1: {}')

        ex = Warning1('Oops', number=1, data={'age': 18})
        self.assertEqual(ex.code, 'Warning1')
        self.assertEqual(ex.message, 'Warning1: Oops')
        self.assertDictEqual(ex.data, {'age': 18})
        self.assertEqual(
            str(ex),
            "[WARN][code|Warning1][message|Warning1: Oops][data|{'age': 18}]"
        )

        self.assertIsInstance(ex, error.Warning)
        self.assertIsInstance(ex, error.BaseEx)
        self.assertNotIsInstance(ex, error.Error)

    def test_manager(self):
        UserError = error.ErrorManager.create_manager('UserError')
        self.assertIn('UserError', error.ErrorManager.managers)

        self.assertDictEqual(UserError.codes, {})
        self.assertDictEqual(UserError.managers, {})

        UserError.add_error('UserNotExist', 'User does not exist')
        self.assertIn('UserError.UserNotExist', UserError.codes)
        self.assertEqual(len(UserError.codes), 1)
        self.assertTrue(issubclass(UserError.UserNotExist, error.Error))
        self.assertTrue(issubclass(UserError.UserNotExist, UserError))

        ex = UserError.UserNotExist(data={'name': 'lucy'})
        self.assertIsInstance(ex, error.Error)
        self.assertIsInstance(ex, UserError)
        self.assertEqual(ex.message, 'User does not exist')
        self.assertDictEqual(ex.data, {'name': 'lucy'})

        with self.assertRaises(error.Error):
            raise ex
        with self.assertRaises(UserError):
            raise ex

        try:
            raise ex
        except UserError.UserNotExist:
            # should catch this exception
            pass

        try:
            raise ex
        except UserError:
            # should catch this exception
            pass

        UserError.add_warning('UserBanned', 'User has been banned')
        self.assertIn('UserError.UserBanned', UserError.codes)
        self.assertEqual(len(UserError.codes), 2)
        self.assertTrue(issubclass(UserError.UserBanned, error.Warning))
        self.assertTrue(issubclass(UserError.UserBanned, UserError))

        ex = UserError.UserBanned(
            data={'reason': 'illegal behaviors', 'until': 2000000000}
        )
        self.assertIsInstance(ex, error.Warning)
        self.assertIsInstance(ex, UserError)
        self.assertEqual(ex.message, 'User has been banned')
        self.assertDictEqual(
            ex.data, {'reason': 'illegal behaviors', 'until': 2000000000}
        )

        with self.assertRaises(error.Warning):
            raise ex
        with self.assertRaises(UserError):
            raise ex

        try:
            raise ex
        except UserError.UserBanned:
            # should catch this exception
            pass

        try:
            raise ex
        except UserError:
            # should catch this exception
            pass

        ProfileError = UserError.create_manager('ProfileError')
        assert ProfileError.__name__ == 'ProfileError'
        self.assertIn('ProfileError', UserError.managers)
        ProfileError.add_warning(
            'IncompleteInfo', 'Please fill in the missing info'
        )

        ex = ProfileError.IncompleteInfo(data={'fields': ['age', 'gender']})
        self.assertIsInstance(ex, error.Warning)
        self.assertIsInstance(ex, ProfileError)
        # yes, ProfileError is managed by UserError
        # and it should be catched by UserError
        self.assertIsInstance(ex, UserError)
        self.assertEqual(ex.code, 'UserError.ProfileError.IncompleteInfo')
        self.assertEqual(ex.message, 'Please fill in the missing info')
        self.assertDictEqual(ex.data, {'fields': ['age', 'gender']})

        with self.assertRaises(error.Warning):
            raise ex
        with self.assertRaises(ProfileError):
            raise ex
        with self.assertRaises(UserError):
            raise ex

        self.assertIsNotNone(
            error.ErrorManager.get_exception('UserError.UserNotExist'),
            error.ErrorManager.codes,
        )
        self.assertIs(
            error.ErrorManager.get_exception('UserError.UserNotExist'),
            UserError.UserNotExist,
        )
        self.assertIsNotNone(
            error.ErrorManager.get_exception(
                'UserError.ProfileError.IncompleteInfo'
            )
        )
        self.assertIs(
            error.ErrorManager.get_exception(
                'UserError.ProfileError.IncompleteInfo'
            ),
            ProfileError.IncompleteInfo,
        )

        assert len(error.ErrorManager.managers) == 1, \
            error.ErrorManager.managers
        assert 'UserError' in error.ErrorManager.managers
        assert 'ProfileError' in UserError.managers

    def test_invalid_action(self):
        UserError = error.ErrorManager.create_manager('Manager')
        UserError.add_error('Error1', 'Manager {} has {number} exceptions')
        with self.assertRaises(ValueError):
            UserError.add_error('Error1', 'Manager {} has {number} exceptions')

        error.ErrorManager.managers.pop('Manager')
        del error.ErrorManager.Manager

    def test_assemble(self):
        class Error1(error.Error):
            code = 'Error1'
            message = 'Error1: {}'
        error.ErrorManager.add('Error1', Error1)
        error.ErrorManager.register(Error1)

        ex = error.ErrorManager.assemble('Error1', 'Error1: assembled', {})
        self.assertIsInstance(ex, Error1)

        ex = error.ErrorManager.assemble(
            'Error2', 'cannot find defination', {}
        )
        self.assertIsInstance(ex, error.Warning)
        self.assertEqual(ex.code, 'Unknown_Error2')
        self.assertEqual(ex.__class__.__name__, 'Unknown_Error2')



from collections import deque
from functools import partial
from typing import Optional, TypeVar, Union

from pymaid.core import get_running_loop
from pymaid.core import Future, TimeoutError
from pymaid.error import BaseEx
from pymaid.utils.logger import logger_wrapper

from .error import RPCError
from .method import Method, MethodStub
from .types import ConnectionType

__all__ = ('C', 'Context', 'InboundContext', 'OutboundContext')


@logger_wrapper
class Context:

    def __init__(
        self,
        *,
        conn: ConnectionType,
        transmission_id: int,
        method: Union[Method, MethodStub],
        timeout: Optional[float] = None,
    ):
        self.conn = conn
        self.conn_id = conn.id
        self.method = method
        self.timeout_interval = timeout
        self.timer = None
        self.waiter = None

        self.transmission_id = transmission_id
        self.is_cancelled = False
        self.is_closed = False
        self.sent_end_message = False

        self.init()

    def init(self):
        pass

    def feed_message(self, message):
        '''Received request from transport layer'''
        raise NotImplementedError('feed_message')

    async def run(self):
        await self.method(self)

    async def shutdown(self):
        pass

    async def close(self, reason: Union[str, Exception] = 'successful'):
        if self.is_closed:
            return
        self.logger.debug(f'{self!r} closed with [reason|{reason}]')
        self.is_closed = True
        if hasattr(self, '_manager'):
            self._manager.release_context(self.transmission_id)
        self.conn = None
        self.method = None
        if self.timer:
            self.timer.cancel()
            self.timer = None
        if self.waiter:
            if reason is None:
                self.waiter.set_result(reason)
            else:
                self.waiter.set_exception(reason)
            self.waiter = None

    async def cancel(self, reason: Optional[Exception] = None):
        self.logger.debug(f'{self!r} cancelled with [reason|{reason}]')
        self.is_cancelled = True
        await self.close(reason)

    async def send_message(
        self, data=None, *, end: bool = False, **kwargs
    ):
        '''Send message to transport layer'''
        raise NotImplementedError('send_message')

    async def __aenter__(self):
        if self.is_closed:
            raise RuntimeError('cannot reuse closed context')
        if self.timeout_interval is not None:
            self.timer = get_running_loop().call_later(
                self.timeout_interval,
                partial(self.cancel, TimeoutError('context action timeout')),
            )
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        await self.close(exc_value)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if m := await self.recv_message():
            return m
        raise StopAsyncIteration()

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} conn={self.conn_id} '
            f'transmission_id={self.transmission_id}>'
        )


class InboundContext(Context):

    def init(self):
        self.request_queue = deque()
        self.request_received_count = 0
        self.request_fed_count = 0
        self.response_sent_count = 0

    async def close(self, reason: Optional[Exception] = None):
        if self.is_closed:
            return
        if isinstance(reason, BaseEx):
            await self.handle_error(reason)
        if self.method.server_streaming and not self.sent_end_message:
            await self.shutdown()
        await super().close(reason)

    async def recv_message(self):
        '''Wait for request from logic layer

        In any cases, this should be called by the same one logic.
        '''
        if (self.request_received_count > 0
                and not self.method.client_streaming):
            raise RPCError.RPCRequestReceived(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )
        if not self.request_queue:
            assert self.waiter is None, \
                'should not called parallelly at the same time'
            self.waiter = Future()
            try:
                await self.waiter
            finally:
                self.waiter = None
        self.request_received_count += 1
        req = self.request_queue.popleft()
        if isinstance(req, Exception):
            raise req
        return req


class OutboundContext(Context):

    def init(self):
        self.request_sent_count = 0
        self.response_queue = deque()
        self.response_received_count = 0
        self.response_fed_count = 0

    async def close(self, reason: Optional[Exception] = None):
        if self.is_closed:
            return
        if isinstance(reason, BaseEx):
            await self.handle_error(reason)
        if self.method.client_streaming and not self.sent_end_message:
            await self.shutdown()
        await super().close(reason)

    async def recv_message(self):
        '''Wait for response from logic layer

        In any cases, this should be called by the same one logic.
        '''
        if self.method.options.get('void_response'):
            return None

        if (self.response_received_count > 0
                and not self.method.server_streaming):
            raise RPCError.RPCResponseReceived(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )
        if not self.response_queue:
            assert self.waiter is None, \
                'should not called parallelly at the same time'
            self.waiter = Future()
            try:
                await self.waiter
            finally:
                self.waiter = None
        self.response_received_count += 1
        resp = self.response_queue.popleft()
        if isinstance(resp, Exception):
            raise resp
        return resp


C = TypeVar('Context', bound=Context)



DEBUG = False

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': (
                u'%(asctime)s.%(msecs).03d %(levelname)s '
                '[pid|%(process)d] %(message)s'
            ),
            'datefmt': '%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
        },
    },
    'loggers': {
        'pymaid': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
        'asyncio': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
    },
}

# Sets the maximum number of consecutive accepts that a process may perform
# on a single wake up. High values give higher priority to high connection
# rates, while lower values give higher priority to already established
# connections.
# Default is 64. Note, that in case of multiple working processes on the
# same listening value, it should be set to a lower value.
MAX_ACCEPT = 64
MAX_TASKS = 32

'''
MAX_CONNECTIONS limits the connections
MAX_CONCURRENCY limits the concurrency of *requests*
MAX_METHOD_CONCURRENCY limits the concurrency of *requests* for specified rpc
e.g.:
1. 10000 connections, next new connection will fail
2. 100 connections, one parallelly call 100 different(not the same) async rpcs
   next rpc call will fail (nomatter from new connection or old connections)
3. 1 connection, parallelly call one async rpc for 10000 times
   next the same rpc call from the connection will fail
'''
MAX_CONNECTIONS = 10000
MAX_CONCURRENCY = 10000
MAX_METHOD_CONCURRENCY = 10000

# connection/socket related settings

PM_KEEPALIVE = True
PM_KEEPIDLE = 60
PM_KEEPINTVL = 5
PM_KEEPCNT = 3

PM_WEBSOCKET_TIMEOUT = 15

SO_SNDBUF = 512 * 1024
SO_RCVBUF = 512 * 1024
SO_SNDTIMEO = 30
SO_RCVTIMEO = 30



from pymaid.core import get_running_loop
from pymaid.ext.middleware import BaseMiddleware

from .error import MonitorError


class HeartbeatMiddleware(BaseMiddleware):

    def __init__(self, heartbeat_interval: int, heartbeat_count: int):
        '''
        :param heartbeat_interval: heartbeat checking interval in seconds.
        :param heartbeat_count:
            heartbeat checking count before marking as heartbeat timeout.
        '''
        self.heartbeat_interval = heartbeat_interval
        self.heartbeat_count = heartbeat_count

    def on_connection_made(self, channel, transport):
        loop = get_running_loop()

        def clear_heartbeat_counter():
            transport.heartbeat_count = 0
            transport.heartbeat_timer.cancel()
            transport.heartbeat_timer = loop.call_later(
                self.heartbeat_timer, heartbeat_timeout,
            )

        def heartbeat_timeout():
            transport.heartbeat_count += 1
            if transport.heartbeat_count >= self.heartbeat_count:
                transport.close(MonitorError.HeartbeatTimeout())
            else:
                transport.heartbeat_timer = loop.call_later(
                    self.heartbeat_interval, heartbeat_timeout,
                )

        transport.heartbeat_count = 0
        transport.heartbeat_timer = loop.call_later(
            self.heartbeat_interval, heartbeat_timeout,
        )
        transport.clear_heartbeat_counter = clear_heartbeat_counter

    def on_connection_lost(self, channel, transport):
        transport.heartbeat_timer.cancel()
        del transport.heartbeat_timer
        del transport.clear_heartbeat_counter
        del transport.heartbeat_count



from __future__ import print_function

import re

from pathlib import Path

from setuptools import setup, find_packages, Extension

root_dir = Path(__file__).parent
__version__ = re.search(
    "__version__ = '(.*)'",
    (root_dir / 'pymaid/__init__.py').read_text(encoding='utf-8'),
    re.M
).group(1)
assert __version__

# Get the long description from the README file
long_description = (root_dir / 'README.md').read_text(encoding='utf-8')

requirements = (root_dir / 'requirements.txt').read_text(encoding='utf-8')
requirements = [line.strip() for line in requirements.split('\n')]

dev_requirements = (root_dir / 'requirements-dev.txt')
dev_requirements = [
    line.strip()
    for line in dev_requirements.read_text(encoding='utf-8').split('\n')
]


if __name__ == '__main__':
    setup(
        name="pymaid",
        description='A rpc framework based on gevent/protobuf',
        long_description=long_description,
        author="Catstyle",
        author_email="Catstyle.Lee@gmail.com",
        url="https://github.com/catstyle/pymaid",
        version=__version__,
        license="GPLv3",

        keywords='asyncio network rpc framework',
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

            # Specify the Python versions you support here.
            # In particular, ensure that you indicate whether you support
            # Python 2, Python 3 or both.
            'Programming Language :: Python :: 3.8',

            'Operating System :: OS Independent',
        ],
        python_requires='>=3.8',

        packages=find_packages(),
        package_data={
            '': ['*.proto'],
        },
        install_requires=requirements,
        tests_require=dev_requirements,
        extras_require={
            'backend': [
                'requests==2.21.0', 'PyYAML==3.13', 'xmltodict==0.12.0'
            ],
        },
        ext_modules=[
            Extension(
                'pymaid.net.ws.speedups',
                sources=['pymaid/net/ws/speedups.c'],
                optional=not (root_dir / '.cibuildwheel').exists(),
            )
        ],
    )



# -----------------------------------------------------------------------------
#   Copyright (C) 2000 Thomas Heller
#   Copyright (C) 2008 Pauli Virtanen <pav@iki.fi>
#   Copyright (C) 2012  The IPython Development Team
#
#   Distributed under the terms of the BSD License.  The full license is in
#   the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------
#
#  This IPython module is written by Pauli Virtanen, based on the autoreload
#  code by Thomas Heller.

# -----------------------------------------------------------------------------
#  Imports
# -----------------------------------------------------------------------------

from __future__ import print_function

import os
import sys
from traceback import format_exc
import types
import weakref
from importlib import import_module

try:
    # Reload is not defined by default in Python3.
    reload
except NameError:
    from imp import reload

try:
    try:
        # Python 3.2, see PEP 3147
        from importlib.util import source_from_cache
    except ImportError:
        # deprecated since 3.4
        from imp import source_from_cache
except ImportError:
    # Python <= 3.1: .pyc files go next to .py
    def source_from_cache(path):
        basename, ext = os.path.splitext(path)
        if ext not in ('.pyc', '.pyo'):
            raise ValueError('Not a cached Python file extension', ext)
        # Should we look for .pyw files?
        return basename + '.py'

if sys.version_info[0] >= 3:
    PY3 = True
else:
    PY3 = False

# ------------------------------------------------------------------------------
# Autoreload functionality
# ------------------------------------------------------------------------------


class ModuleReloader(object):

    enabled = False
    """Whether this reloader is enabled"""

    check_all = True
    """Autoreload all modules, not just those listed in 'modules'"""

    def __init__(self):
        # Modules that failed to reload: {module: mtime-on-failed-reload, ...}
        self.failed = {}
        # Modules specially marked as autoreloadable.
        self.modules = {}
        # Modules specially marked as not autoreloadable.
        self.skip_modules = {}
        # (module-name, name) -> weakref, for replacing old code objects
        self.old_objects = {}
        # Module modification timestamps
        self.modules_mtimes = {}

        # Cache module modification times
        self.check(check_all=True, do_reload=False)

    def mark_module_skipped(self, module_name):
        """Skip reloading the named module in the future"""
        try:
            del self.modules[module_name]
        except KeyError:
            pass
        self.skip_modules[module_name] = True

    def mark_module_reloadable(self, module_name):
        """Reload the named module in the future (if it is imported)"""
        try:
            del self.skip_modules[module_name]
        except KeyError:
            pass
        self.modules[module_name] = True

    def aimport_module(self, module_name):
        """Import a module, and mark it reloadable

        Returns
        -------
        top_module : module
            The imported module if it is top-level, or the top-level
        top_name : module
            Name of top_module

        """
        self.mark_module_reloadable(module_name)

        import_module(module_name)
        top_name = module_name.split('.')[0]
        top_module = sys.modules[top_name]
        return top_module, top_name

    def filename_and_mtime(self, module):
        if not hasattr(module, '__file__') or module.__file__ is None:
            return None, None

        if getattr(module, '__name__', None) == '__main__':
            # we cannot reload(__main__)
            return None, None

        filename = module.__file__
        path, ext = os.path.splitext(filename)

        if ext.lower() == '.py':
            py_filename = filename
        else:
            try:
                py_filename = source_from_cache(filename)
            except ValueError:
                return None, None

        try:
            pymtime = os.stat(py_filename).st_mtime
        except OSError:
            return None, None

        return py_filename, pymtime

    def check(self, check_all=False, do_reload=True):
        """Check whether some modules need to be reloaded."""

        if not self.enabled and not check_all:
            return

        if check_all or self.check_all:
            modules = list(sys.modules.keys())
        else:
            modules = list(self.modules.keys())

        for modname in modules:
            m = sys.modules.get(modname, None)

            if modname in self.skip_modules:
                continue

            py_filename, pymtime = self.filename_and_mtime(m)
            if py_filename is None:
                continue

            try:
                if pymtime <= self.modules_mtimes[modname]:
                    continue
            except KeyError:
                self.modules_mtimes[modname] = pymtime
                continue
            else:
                if self.failed.get(py_filename, None) == pymtime:
                    continue

            self.modules_mtimes[modname] = pymtime

            # If we've reached this point, we should try to reload the module
            if do_reload:
                try:
                    superreload(m, reload, self.old_objects)
                    if py_filename in self.failed:
                        del self.failed[py_filename]
                except Exception:
                    print(
                        f"[autoreload of {modname} failed: {format_exc()}]",  # noqa
                        file=sys.stderr
                    )
                    self.failed[py_filename] = pymtime

# ------------------------------------------------------------------------------
# superreload
# ------------------------------------------------------------------------------


if PY3:
    func_attrs = ['__code__', '__defaults__', '__doc__',
                  '__closure__', '__globals__', '__dict__']
else:
    func_attrs = ['func_code', 'func_defaults', 'func_doc',
                  'func_closure', 'func_globals', 'func_dict']


def update_function(old, new):
    """Upgrade the code object of a function"""
    for name in func_attrs:
        try:
            setattr(old, name, getattr(new, name))
        except (AttributeError, TypeError):
            pass


def update_class(old, new):
    """Replace stuff in the __dict__ of a class, and upgrade
    method code objects"""
    for key in list(old.__dict__.keys()):
        old_obj = getattr(old, key)

        try:
            new_obj = getattr(new, key)
        except AttributeError:
            # obsolete attribute: remove it
            try:
                delattr(old, key)
            except (AttributeError, TypeError):
                pass
            continue

        if update_generic(old_obj, new_obj):
            continue

        try:
            setattr(old, key, getattr(new, key))
        except (AttributeError, TypeError):
            pass  # skip non-writable attributes


def update_property(old, new):
    """Replace get/set/del functions of a property"""
    update_generic(old.fdel, new.fdel)
    update_generic(old.fget, new.fget)
    update_generic(old.fset, new.fset)


def isinstance2(a, b, typ):
    return isinstance(a, typ) and isinstance(b, typ)


UPDATE_RULES = [
    (lambda a, b: isinstance2(a, b, type),
     update_class),
    (lambda a, b: isinstance2(a, b, types.FunctionType),
     update_function),
    (lambda a, b: isinstance2(a, b, property),
     update_property),
]


if PY3:
    UPDATE_RULES.extend([
        (lambda a, b: isinstance2(a, b, types.MethodType),
         lambda a, b: update_function(a.__func__, b.__func__)),
    ])
else:
    UPDATE_RULES.extend([
        (lambda a, b: isinstance2(a, b, types.ClassType),
         update_class),
        (lambda a, b: isinstance2(a, b, types.MethodType),
         lambda a, b: update_function(a.__func__, b.__func__)),
    ])


def update_generic(a, b):
    for type_check, update in UPDATE_RULES:
        if type_check(a, b):
            update(a, b)
            return True
    return False


class StrongRef(object):

    def __init__(self, obj):
        self.obj = obj

    def __call__(self):
        return self.obj


def superreload(module, reload=reload, old_objects={}):
    """Enhanced version of the builtin reload function.

    superreload remembers objects previously in the module, and

    - upgrades the class dictionary of every old class in the module
    - upgrades the code object of every old function and method
    - clears the module's namespace before reloading

    """

    # collect old objects in the module
    for name, obj in list(module.__dict__.items()):
        if not hasattr(obj, '__module__') or obj.__module__ != module.__name__:
            continue
        key = (module.__name__, name)
        try:
            old_objects.setdefault(key, []).append(weakref.ref(obj))
        except TypeError:
            # weakref doesn't work for all types;
            # create strong references for 'important' cases
            if not PY3 and isinstance(obj, types.ClassType):
                old_objects.setdefault(key, []).append(StrongRef(obj))

    # reload module
    try:
        # clear namespace first from old cruft
        old_dict = module.__dict__.copy()
        old_name = module.__name__
        module.__dict__.clear()
        module.__dict__['__name__'] = old_name
        module.__dict__['__loader__'] = old_dict['__loader__']
    except (TypeError, AttributeError, KeyError):
        pass

    try:
        module = reload(module)
    except Exception:
        # restore module dictionary on failed reload
        module.__dict__.update(old_dict)
        raise

    # iterate over all objects and update functions & classes
    for name, new_obj in list(module.__dict__.items()):
        key = (module.__name__, name)
        if key not in old_objects:
            continue

        new_refs = []
        for old_ref in old_objects[key]:
            old_obj = old_ref()
            if old_obj is None:
                continue
            new_refs.append(old_ref)
            update_generic(old_obj, new_obj)

        if new_refs:
            old_objects[key] = new_refs
        else:
            del old_objects[key]

    return module


def enable_autoreload(signum):
    reloader = ModuleReloader()
    reloader.enabled = True

    def autoreload(sig, frame):
        reloader.check()

    from pymaid import backend
    backend.get_running_loop().add_signal_handler(signum, autoreload)



import abc

from typing import Any, TypeVar

from pymaid.types import DataType


class Protocol(abc.ABC):
    '''pymaid use Protocol representation app protocol layer

    You can build your protocol upon Protocol
    and you can easily change the underlying Protocol
    e.g.:
        build your AppProtocol inherit from Http and change to Http2 if wanted,
        and in best case, do not need to do anything else

        class AppProtocol(Http):
            ...

        class AppProtocol(Http2):
            ...
    '''

    @abc.abstractclassmethod
    def feed_data(cls, data: DataType):
        raise NotImplementedError

    @abc.abstractclassmethod
    def encode(cls, obj: Any) -> DataType:
        raise NotImplementedError

    @abc.abstractclassmethod
    def decode(cls, data: DataType) -> Any:
        raise NotImplementedError


ProtocolType = TypeVar('Protocol', bound=Protocol)



import abc
from typing import AsyncIterable, Callable, Optional, Type

from .types import ConnectionType, InboundContext, Request, Response


class Method(abc.ABC):

    def __init__(
        self,
        name: str,
        full_name: str,
        method_impl: Callable,
        request_class: Type[Request],
        response_class: Type[Response],
        *,
        options: dict = None,
    ):
        self.name = name
        self.full_name = full_name
        self.method_impl = method_impl
        self.request_class = request_class
        self.response_class = response_class
        self.options = options or {}

    async def __call__(self, context: InboundContext):
        async with context:
            await self.method_impl(context)


class UnaryUnaryMethod(Method):

    client_streaming = False
    server_streaming = False


class UnaryStreamMethod(Method):

    client_streaming = False
    server_streaming = True


class StreamUnaryMethod(Method):

    client_streaming = True
    server_streaming = False


class StreamStreamMethod(Method):

    client_streaming = True
    server_streaming = True


class MethodStub(metaclass=abc.ABCMeta):

    def __init__(
        self,
        name: str,
        full_name: str,
        request_class: Type[Request],
        response_class: Type[Response],
        *,
        options: dict = None,
    ):
        self.name = name
        self.full_name = full_name
        self.request_class = request_class
        self.response_class = response_class
        self.options = options or {}

    def open(
        self,
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        return conn.handler.new_outbound_context(method=self, timeout=timeout)

    @abc.abstractmethod
    def __call__(
        self,
        request: Request,
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        raise NotImplementedError()


class UnaryUnaryMethodStub(MethodStub):

    client_streaming = False
    server_streaming = False

    async def __call__(
        self,
        request: Request,
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        async with self.open(conn=conn, timeout=timeout) as context:
            await context.send_message(request)
            return await context.recv_message()


class UnaryStreamMethodStub(MethodStub):

    client_streaming = False
    server_streaming = True

    async def __call__(
        self,
        request: Request,
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        async with self.open(conn=conn, timeout=timeout) as context:
            await context.send_message(request)
            async for resp in context:
                yield resp


class StreamUnaryMethodStub(MethodStub):

    client_streaming = True
    server_streaming = False

    async def __call__(
        self,
        requests: AsyncIterable[Request],
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        async with self.open(conn=conn, timeout=timeout) as context:
            async for request in requests:
                await context.send_message(request)
            await context.send_message(end=True)
            return await context.recv_message()


class StreamStreamMethodStub(MethodStub):

    client_streaming = True
    server_streaming = True

    async def __call__(
        self,
        requests: AsyncIterable[Request],
        *,
        conn: ConnectionType,
        timeout: Optional[float] = None,
    ):
        async with self.open(conn=conn, timeout=timeout) as context:
            async for request in requests:
                await context.send_message(request)
            await context.send_message(end=True)
            async for resp in context:
                yield resp



from functools import wraps
from unittest import mock

import pytest

from pymaid.core import iscoroutinefunction
from pymaid.utils.functional import with_defer, defer


def not_wrapped(func):
    return func


def wrapped(func):
    if iscoroutinefunction(func):
        @wraps(func)
        async def _(*args, **kwargs):
            return await func(*args, **kwargs)
        return _
    else:
        @wraps(func)
        def _(*args, **kwargs):
            return func(*args, **kwargs)
        return _


def test_defer():
    m = mock.Mock()

    @with_defer
    def func():
        defer(m, 'called by defer')

    func()
    m.assert_called_once_with('called by defer')

    m = mock.Mock()

    @with_defer
    def func(count):
        for idx in range(count):
            defer(m, idx)

    func(10)
    assert m.call_count == 10
    args = [c.args[0] for c in m.call_args_list]
    assert args == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]


@pytest.mark.asyncio
async def test_defer_async():
    am = mock.AsyncMock()

    @with_defer
    async def func():
        defer(am, 'called by defer')

    await func()
    am.assert_called_once_with('called by defer')

    # async and sync comb
    am = mock.AsyncMock()
    m = mock.Mock()

    @with_defer
    async def func(count):
        for idx in range(count):
            defer(am, idx)
        defer(m, 'called by defer')

    await func(10)
    assert am.call_count == 10
    args = [c.args[0] for c in am.call_args_list]
    assert args == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    m.assert_called_once_with('called by defer')


def test_no_with_defer():
    def func():
        defer(print, 'should not happen')

    # no decorator
    with pytest.raises(RuntimeError):
        func()


def test_sync_function_defer_async():
    am = mock.AsyncMock()

    @with_defer
    def func():
        defer(am, 'called by defer')

    # sync function but defer async
    with pytest.raises(ValueError):
        func()


@pytest.mark.asyncio
async def test_defer_not_callable():
    @with_defer
    def func():
        defer(1, 'called by defer')

    # sync function but defer not callable
    with pytest.raises(TypeError):
        func()

    @with_defer
    async def func():
        defer(1, 'called by defer')

    # async function but def not callable
    with pytest.raises(TypeError):
        await func()


@pytest.mark.asyncio
async def test_invalid_defer_scope():
    m = mock.Mock()

    @with_defer
    def func():
        def inner():
            defer(m, 'called by defer')
        inner()

    # defer from inner
    with pytest.raises(RuntimeError):
        func()

    am = mock.AsyncMock()

    @with_defer
    async def func():
        def inner():
            defer(am, 'called by defer')
        inner()

    # defer from inner
    with pytest.raises(RuntimeError):
        await func()


@pytest.mark.asyncio
async def test_multiple_decorators():

    m = mock.Mock()

    @not_wrapped
    @with_defer
    def func():
        defer(m, 'called by defer')
    func()
    m.assert_called_once_with('called by defer')

    # NOTICE, if decorator donot wraps the function calling, its order does not matter
    m = mock.Mock()

    @with_defer
    @not_wrapped
    def func():
        defer(m, 'called by defer')
    func()
    m.assert_called_once_with('called by defer')

    m = mock.Mock()

    @wrapped
    @with_defer
    def func():
        defer(m, 'called by defer')
    func()
    m.assert_called_once_with('called by defer')

    am = mock.Mock()

    @not_wrapped
    @with_defer
    async def func():
        defer(am, 'called by defer')
    await func()
    am.assert_called_once_with('called by defer')

    # NOTICE, if a decorator donot wraps the function calling, its order does not matter
    am = mock.Mock()

    @with_defer
    @not_wrapped
    async def func():
        defer(am, 'called by defer')
    await func()
    am.assert_called_once_with('called by defer')

    am = mock.Mock()

    @wrapped
    @with_defer
    async def func():
        defer(am, 'called by defer')
    await func()
    am.assert_called_once_with('called by defer')


@pytest.mark.asyncio
async def test_multiple_decorators_invalid_usage():
    m = mock.Mock()

    @with_defer
    @wrapped
    def func():
        defer(m, 'called by defer')

    # with_defer not the innermost one
    with pytest.raises(RuntimeError):
        func()

    am = mock.Mock()

    @with_defer
    @wrapped
    async def func():
        defer(am, 'called by defer')

    # with_defer not the innermost one
    with pytest.raises(RuntimeError):
        await func()



from socket import error as socket_error


class WebSocketError(socket_error):
    """Base class for all websocket errors."""


class ProtocolError(WebSocketError):
    """Raised if an error occurs when de/encoding the websocket protocol."""


class FrameTooLargeException(ProtocolError):
    """Raised if a frame is received that is too large."""



import socket
import ssl as _ssl

from typing import Callable, List, Optional, TypeVar

from pymaid.conf import settings
from pymaid.types import DataType

from .transport import SocketTransport


class Stream(SocketTransport):

    MAX_SIZE = 256 * 1024  # recv size passed to sock.recv
    KEEP_OPEN_ON_EOF = False

    def __init__(
        self,
        sock: socket.socket,
        *,
        on_open: Optional[List[Callable]] = None,
        on_close: Optional[List[Callable]] = None,
        initiative: bool = False,
        ssl_context: Optional[_ssl.SSLContext] = None,
        ssl_handshake_timeout: Optional[float] = None,
    ):
        super().__init__(sock, on_open=on_open, on_close=on_close)
        self.initiative = initiative
        self.ssl_context = ssl_context
        self.ssl_handshake_timeout = ssl_handshake_timeout
        self._write_empty_waiter = None

        for cb in self.on_open:
            cb(self)

    def set_socket_default_options(self):
        super().set_socket_default_options()
        sock = self._sock
        assert sock.type == socket.SOCK_STREAM

        setsockopt = sock.setsockopt
        getsockopt = sock.getsockopt
        SOL_SOCKET, SOL_TCP = socket.SOL_SOCKET, socket.SOL_TCP

        if sock.family != socket.AF_UNIX:
            setsockopt(SOL_TCP, socket.TCP_NODELAY, 1)
            setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        ns = settings.namespaces['pymaid']  # should always exists
        if getsockopt(SOL_SOCKET, socket.SO_SNDBUF) < ns['SO_SNDBUF']:
            setsockopt(SOL_SOCKET, socket.SO_SNDBUF, ns['SO_SNDBUF'])
        if getsockopt(SOL_SOCKET, socket.SO_RCVBUF) < ns['SO_RCVBUF']:
            setsockopt(SOL_SOCKET, socket.SO_RCVBUF, ns['SO_RCVBUF'])

        if sock.family != socket.AF_UNIX and ns['PM_KEEPALIVE']:
            setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            setsockopt(SOL_TCP, socket.TCP_KEEPIDLE, ns['PM_KEEPIDLE'])
            setsockopt(SOL_TCP, socket.TCP_KEEPINTVL, ns['PM_KEEPINTVL'])
            setsockopt(SOL_TCP, socket.TCP_KEEPCNT, ns['PM_KEEPCNT'])

    def write_sync(self, data: DataType) -> bool:
        '''Write data to low level socket, in a synchronized way.

        Do some optimization by sending data directly if possible.
        Otherwise will add data to write_buffer, do it in write io Callable.

        There is another :meth:`write`, it will take care of write_buffer.

        :returns: bool, indicate whether sent out all data this time or not.
        '''
        if not self.write_buffer:
            # Optimization: try to send now.
            try:
                n = self._sock.send(data)
            except (BlockingIOError, InterruptedError):
                pass
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as exc:
                self._fatal_error(exc, 'Fatal write error on socket transport')
                return
            else:
                data = data[n:]
                if not data:
                    return True
            # Not all was written; register write handler.
            self._loop.add_writer(self._sock_fd, self._write_ready)

        # Add it to the buffer.
        self.write_buffer.extend(data)
        return False

    async def write(self, data: DataType):
        '''Write data to low level socket, in an asynchronized way.

        Do some optimization by sending data directly if possible.
        Otherwise will add data to write_buffer, do it in write io Callable.

        In order to deal with the issue of `handle backpressure correctly`_
        Will try to call await on the :meth:`wait_write_all` to wait for all
        buffered data to send.

        .. _handle backpressure correctly: https://vorpus.org/blog/some-thoughts-on-asynchronous-api-design-in-a-post-asyncawait-world/#bug-1-backpressure  # noqa
        '''
        if not self.write_sync(data):
            await self.wait_write_all()

    async def wait_write_all(self, timeout=None):
        '''Wait for all buffered data to send.

        Note: data added to the buffer during this period will be waited too.

        :params timeout: If passed an int or float, expressed in seconds, will
            cancel this method if sending not finished after that time. It is
            always relative to the current time.
        '''
        if not self.write_buffer:
            return
        if self._write_empty_waiter is not None:
            raise RuntimeError('wait_write_all has been called')
        self._write_empty_waiter = self._loop.create_future()
        if timeout is not None:
            timer = self._loop.call_later(
                timeout, self._write_empty_waiter.cancel
            )
        try:
            await self._write_empty_waiter
        finally:
            if timeout is not None:
                timer.cancel()
            self._write_empty_waiter = None

    def data_received(self, data: DataType):
        self.logger.debug(f'{self!r} data_received, {len(data)=}, ignored!')

    def eof_received(self) -> bool:
        '''Returned value indicate whether to keep the transport open or not.

        Default will close the transport.
        '''
        self.logger.debug(f'{self!r} eof_received')
        return self.KEEP_OPEN_ON_EOF

    def _read_ready(self):
        try:
            data = self._sock.recv(self.MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            return
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            self._fatal_error(exc, 'Fatal read error on socket transport')
            return

        if not data:
            try:
                if self.eof_received():
                    # We're keeping the connection open so the
                    # protocol can write more, but we still can't
                    # receive more, so remove the reader Callable.
                    self._loop.remove_reader(self._sock_fd)
                else:
                    self.close()
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as exc:
                self._fatal_error(exc, 'Fatal error: eof_received() failed.')
            return

        try:
            self.data_received(data)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            self._fatal_error(
                exc, 'Fatal error: stream.data_received() call failed.'
            )

    def _write_ready(self):
        assert self.write_buffer, 'data should not be empty'

        try:
            n = self._sock.send(self.write_buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            self._fatal_error(exc, 'Fatal write error on socket transport')
        else:
            if n:
                del self.write_buffer[:n]
            if not self.write_buffer:
                self._loop.remove_writer(self._sock_fd)
                if self._write_empty_waiter:
                    self._write_empty_waiter.set_result(None)
                    self._write_empty_waiter = None
                if self.state == self.STATE.CLOSING:
                    self._finnal_close(None)


StreamType = TypeVar('StreamType', bound=Stream)



from pymaid.net.channel import StreamChannel
from pymaid.net.stream import Stream
from pymaid.types import DataType


class _TestStreamChannel(StreamChannel):

    # for test case only
    def connection_made(self, sock):
        self.connected_stream = super().connection_made(sock)
        return self.connected_stream


class _TestStream(Stream):

    # for test case only
    def data_received(self, data: DataType):
        self.logger.debug(f'{self!r} data_received, {len(data)=}')
        self.received_data = data


del StreamChannel, Stream



import ssl as _ssl
from typing import Optional, Sequence, Tuple, Type, Union

from pymaid.ext.middleware import MiddlewareManager
from pymaid.net.protocol import Protocol, ProtocolType

from . import channel
from . import connection
from . import handler
from . import pb
from . import router

from .types import ServiceType

__all__ = ('channel', 'connection', 'handler', 'pb', 'router')


async def serve_stream(
    address: Union[Tuple[str, int], str],
    *,
    name: str = 'StreamChannel',
    transport_class: connection.ConnectionType = connection.Connection,
    backlog: int = 128,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    middleware_manager: Optional[MiddlewareManager] = None,
    close_conn_onerror: bool = True,
    protocol: Type[ProtocolType] = Protocol,
    handler_class: Optional[Type[handler.Handler]] = handler.SerialHandler,
    services: Optional[Sequence[ServiceType]] = None,
    router: Optional[router.Router] = None,
):
    assert services is not None or router, 'should provide services or router'
    ch = channel.create_stream_channel(
        address,
        name=name,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        middleware_manager=middleware_manager,
        close_conn_onerror=close_conn_onerror,
        protocol=protocol,
        handler_class=handler_class,
    )
    if services:
        ch.router.include_services(services)
    if router:
        ch.router.include_router(router)
    await ch.listen(address, backlog=backlog)
    ch.start()
    async with ch:
        await ch.serve_forever()



import pymaid.rpc.pb

from examples.template import get_client_parser, parse_args


async def wrapper(ch):
    conn = await ch.acquire()
    # will be closed by server due to heartbeat timeout
    await conn.wait_closed()


async def main(args):
    ch = await pymaid.rpc.pb.dial_stream(args.address)
    tasks = []
    for x in range(args.concurrency):
        tasks.append(pymaid.create_task(wrapper(ch)))

    # await pymaid.wait(tasks, timeout=args.timeout)
    await pymaid.gather(*tasks)


if __name__ == '__main__':
    args = parse_args(get_client_parser())
    pymaid.run(main(args))



import traceback

from io import BytesIO

import httptools

from . import Protocol


CRITICAL_ERROR_TEXT = '''HTTP/1.0 500 INTERNAL SERVER ERROR
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Type: text/html; charset=utf-8
Content-Length: 127

<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 3.2 Final//EN'>
<title>Internal Server Error</title>
<h1>Internal Server Error</h1>
'''.replace('\n', '\r\n')

HTTP_SWITCHING_PROTOCOLS = '''HTTP/1.1 101 SWITCHING PROTOCOLS
Connection: Upgrade
Upgrade: h2c
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 0

'''.replace('\n', '\r\n')

HTTP_TOO_BIG = '''HTTP/1.1 413 PAYLOAD TOO LARGE
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 0

'''.replace('\n', '\r\n')

HTTP_INVALID_COMPRESSION = '''HTTP/1.1 400 BAD REQUEST
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 25

Invalid compressed data
'''.replace('\n', '\r\n')


class HttpParser:

    MAX_BODY_SIZE = 10 * 1024 * 1024

    def __init__(self, transport, handler):
        self.transport = transport
        self.handler = handler

    # httptools callbacks
    def on_message_begin(self):
        ''' Called when a message begins. '''
        self.body = BytesIO()
        self.headers = []

    def on_header(self, name: bytes, value: bytes):
        ''' Called when a header has been received.

        :param name: The name of the header.
        :param value: The value of the header.
        '''
        self.headers.append((name.decode(), value.decode()))

    def on_headers_complete(self):
        ''' Called when the headers have been completely sent. '''

    def on_body(self, body: bytes):
        ''' Called when part of the body has been received.

        :param body: The body text.
        '''
        self.body.write(body)
        if self.body.tell() >= self.MAX_BODY_SIZE:
            # write a 'too big' message
            self.write(HTTP_TOO_BIG)
            self.close()

    def on_chunk_header(self):
        pass

    def on_chunk_complete(self):
        pass


class HttpRequest(HttpParser):

    def on_message_begin(self):
        super().on_message_begin()
        self.full_url = ''

    def on_url(self, url: bytes):
        self.full_url = url.decode('utf-8')

    def on_message_complete(self):
        self.handler.handle_request(self)


class HttpResponse(HttpParser):

    def on_message_begin(self):
        super().on_message_begin()
        self.status = 0

    def on_status(self, url: bytes):
        self.full_url = url.decode('utf-8')

    def on_message_complete(self):
        self.handler.handle_response(self)


class HTTP(Protocol):

    # request_parser = httptools.HttpRequestParser(HttpRequest())
    # response_parser = httptools.HttpResponseParser(HttpResponse())

    @classmethod
    def feed_data(cls, data: bytes):
        try:
            cls.parser.feed_data(data)
        except httptools.HttpParserUpgrade:
            cls.handle_upgrade()
        except httptools.HttpParserError as exc:
            traceback.print_exc()
            cls.handle_parser_exception(exc)

    @classmethod
    def handle_parser_exception(cls, exc):
        pass

    @classmethod
    def handle_upgrade(cls):
        pass



import os
import socket

import pytest

from pymaid.net.raw import sock_connect, sock_listen
from pymaid.net.raw import HAS_IPv6_FAMILY


@pytest.mark.asyncio
async def test_sock_listen_ipv4():
    sockets = await sock_listen(('localhost', 8999))
    assert sockets
    for sock in sockets:
        assert sock.family == socket.AF_INET
        assert sock.type == socket.SOCK_STREAM
        sock.close()


@pytest.mark.skipif(not HAS_IPv6_FAMILY, reason='does not support ipv6')
@pytest.mark.asyncio
async def test_sock_listen_ipv6():
    sockets = await sock_listen(('::1', 8999))
    assert sockets
    for sock in sockets:
        assert sock.family == socket.AF_INET6
        assert sock.type == socket.SOCK_STREAM
        sock.close()


@pytest.mark.skipif(
    not getattr(socket, 'AF_UNIX', None),
    reason='does not support unix domain sock'
)
@pytest.mark.asyncio
async def test_sock_listen_unix():
    sockets = await sock_listen('/tmp/pymaid_test.sock')
    assert sockets
    for sock in sockets:
        assert sock.family == socket.AF_UNIX
        assert sock.type == socket.SOCK_STREAM
        sock.close()
    os.unlink('/tmp/pymaid_test.sock')


@pytest.mark.asyncio
async def test_sock_connect_ipv4():
    sockets = await sock_listen(('localhost', 8999))
    assert sockets

    sock = await sock_connect(('localhost', 8999))
    assert sock.family == socket.AF_INET
    # localhost resolved to 127.0.0.1
    assert sock.getpeername() == ('127.0.0.1', 8999)

    for listen_sock in sockets:
        peer, addr = listen_sock.accept()
        break
    sock.send(b'from pymaid')
    assert peer.recv(1024) == b'from pymaid'

    peer.close()
    for sock in sockets:
        sock.close()


@pytest.mark.skipif(not HAS_IPv6_FAMILY, reason='does not support ipv6')
@pytest.mark.asyncio
async def test_sock_connect_ipv6():
    sockets = await sock_listen(('::1', 8999))
    assert sockets
    assert len(sockets) == 1

    sock = await sock_connect(('::1', 8999))
    assert sock.family == socket.AF_INET6
    assert sock.getpeername()[:2] == ('::1', 8999)

    for listen_sock in sockets:
        peer, addr = listen_sock.accept()
        break
    sock.send(b'from pymaid')
    assert peer.recv(1024) == b'from pymaid'

    peer.close()
    for sock in sockets:
        sock.close()


@pytest.mark.skipif(
    not getattr(socket, 'AF_UNIX', None),
    reason='does not support unix domain sock'
)
@pytest.mark.asyncio
async def test_sock_connect_unix():
    sockets = await sock_listen('/tmp/pymaid_test_ipv6.sock')
    assert sockets
    assert len(sockets) == 1

    sock = await sock_connect('/tmp/pymaid_test_ipv6.sock')
    assert sock.family == socket.AF_UNIX
    assert sock.getpeername() == '/tmp/pymaid_test_ipv6.sock'

    for listen_sock in sockets:
        peer, addr = listen_sock.accept()
        break
    sock.send(b'from pymaid')
    assert peer.recv(1024) == b'from pymaid'

    peer.close()
    for sock in sockets:
        sock.close()

    os.unlink('/tmp/pymaid_test_ipv6.sock')



from orjson import loads, dumps

from pymaid.core import create_task
from pymaid.utils.logger import logger_wrapper


def formatter(*formats):
    def wrapper(func):
        func.formats = formats
        return func
    return wrapper


class BackendMeta(type):

    def __new__(cls, name, bases, attrs):
        cls = type.__new__(cls, name, bases, attrs)
        cls.formatters = {}
        for attr in cls.__dict__.keys():
            attr = getattr(cls, attr)
            for format in getattr(attr, 'formats', []):
                cls.formatters[format] = attr
        return cls


class SettingsBackend(metaclass=BackendMeta):

    def __init__(self):
        self.listening = False
        self.fetching = False
        self.subscriptions = {}

    def subscribe(self, ns, settings, format='json'):
        if format not in self.formatters:
            self.logger.warn(
                '[pymaid][settings|%s] subscribed with unknown [format|%s]',
                settings, format
            )
            return False

        if ns in self.subscriptions:
            self.logger.warn(
                '[pymaid][settings|%s] already subscribed with [ns|%s]',
                settings, ns
            )
            return False
        else:
            self.subscriptions[ns] = {'settings': settings, 'format': format}
            return True

    def start(self):
        self.listening = True
        self.worker = create_task(self.run())

    async def run(self):
        raise NotImplementedError('run')

    def stop(self):
        self.listening = False
        if hasattr(self, 'worker'):
            self.worker.kill()


@logger_wrapper
class ApolloBackend(SettingsBackend):

    def __init__(self,
                 app_id,
                 cluster='default',
                 config_server='http://localhost:8080',
                 timeout=65,
                 api_type='cached'):
        super(ApolloBackend, self).__init__()
        self.config_server = config_server
        self.app_id = app_id
        self.cluster = cluster
        self.timeout = timeout
        self.api_type = api_type
        import requests
        self.session = requests.Session()

    def subscribe(self, ns, settings, format='json'):
        result = super(ApolloBackend, self).subscribe(ns, settings, format)
        if result:
            self.subscriptions[ns]['notificationId'] = -1
            if self.listening and not self.fetching:
                self.stop()  # stop terminate listerner
                self.start()  # start listerner again
        return result

    async def run(self):
        import requests
        while self.listening:
            notifications = [
                {
                    'namespaceName': f'{ns}.{sub["format"]}',
                    'notificationId': sub['notificationId'],
                }
                for ns, sub in self.subscriptions.items()
            ]
            try:
                resp = self.session.get(
                    f'{self.config_server}/notifications/v2',
                    params={
                        'appId': self.app_id,
                        'cluster': self.cluster,
                        'notifications': dumps(notifications),
                    },
                    timeout=self.timeout,
                )
            except requests.exceptions.RequestException as ex:
                self.logger.warn(
                    '[pymaid][settings][backend|%s] raise request error: %r',
                    self.__class__.__name__, ex
                )
                continue

            if resp.status_code == 304:
                self.logger.debug(
                    '[pymaid][settings][backend|%s] no changes, continue',
                    self.__class__.__name__,
                )
                continue

            if resp.ok:
                update = resp.json()
                self.logger.debug(
                    '[pymaid][settings][backend|%s] receive [update|%r]',
                    self.__class__.__name__, update
                )
                self.fetching = True
                delta = {}
                if self.api_type == 'cached':
                    get_data = self.get_cached_data
                else:
                    get_data = self.get_uncached_data
                for item in update:
                    ns = item['namespaceName'].rsplit('.', 1)[0]
                    nid = item['notificationId']
                    data = get_data(ns, self.subscriptions[ns]['format'])
                    if data:
                        self.subscriptions[ns]['notificationId'] = nid
                        delta[ns] = data

                for ns, data in delta.items():
                    settings = self.subscriptions[ns]['settings']
                    settings.load_from_object(data, ns)
                self.logger.debug(
                    '[pymaid][settings][backend|%s] updated [delta|%r]',
                    self.__class__.__name__, delta.keys()
                )
                self.fetching = False

    def get_cached_data(self, ns, format):
        import requests
        try:
            resp = self.session.get(
                f'{self.config_server}/configfiles/json/'
                f'{self.app_id}/{self.cluster}/{ns}.{format}',
                timeout=self.timeout,
            )
        except requests.exceptions.RequestException as ex:
            self.logger.warn(
                '[pymaid][settings][backend|%s] raise request error: %r',
                self.__class__.__name__, ex
            )
        else:
            if resp.ok:
                return ApolloBackend.formatters[format](self, resp.json())

    def get_uncached_data(self, ns, format):
        import requests
        try:
            resp = self.session.get(
                f'{self.config_server}/configs/'
                f'{self.app_id}/{self.cluster}/{ns}.{format}',
                timeout=self.timeout,
            )
        except requests.exceptions.RequestException as ex:
            self.logger.warn(
                '[pymaid][settings][backend|%s] raise request error: %r',
                self.__class__.__name__, ex
            )
        else:
            if resp.ok:
                return ApolloBackend.formatters[format](
                    self, resp.json()['configurations']
                )

    @formatter('json')
    def json_formatter(self, content):
        return loads(content['content'])

    @formatter('properties')
    def properties_formatter(self, content):
        return content

    @formatter('yaml', 'yml')
    def yaml_formatter(self, content):
        import yaml
        return yaml.load(content['content'])

    @formatter('xml')
    def xml_formatter(self, content):
        import xmltodict
        return xmltodict.parse(content['content'])



from typing import TypeVar

ConnectionType = TypeVar('Connection')

Request = TypeVar('Request')
Response = TypeVar('Response')

Context = TypeVar('Context')
InboundContext = TypeVar('InboundContext', bound=Context)
OutboundContext = TypeVar('OutboundContext', bound=Context)
Method = TypeVar('Method')
MethodStub = TypeVar('MethodStub')

RouterType = TypeVar('Router')
ServiceType = TypeVar('Service')
HandlerType = TypeVar('Handler')



from pymaid.rpc.pb import implall

from .monitor_pb2 import MonitorService, Pong


@implall
class MonitorServiceImpl(MonitorService):

    async def Ping(self, context) -> Pong:
        if not context.conn.is_closed:
            context.conn.clear_heartbeat_counter()
        await context.send_message(Pong())
